from treelib import Tree, Node
import time
import seqpy
from Bio import SeqIO
import numpy as np
from collections import defaultdict
import bidict
import random
import os


# hierarchical clustering
def hierarchy(fna_mapping, dist):
    pending = list(fna_mapping.keys())
    node_id = max(pending) + 1
    mapping = bidict.bidict()
    cls_dist = []
    cls_dist_temp = {}
    index = 0
    pending.sort()
    for i in pending:
        mapping[i] = index
        index += 1
    for i in range(0, len(pending)):
        temp1 = []
        for j in range(0, i):
            temp1.append(cls_dist_temp[(mapping[pending[j]], mapping[pending[i]])])
        for j in range(i, len(pending)):
            temp = cal_cls_dist(dist, fna_mapping[pending[i]], fna_mapping[pending[j]])
            temp1.append(temp)
            cls_dist_temp[(mapping[pending[i]], mapping[pending[j]])] = temp
        cls_dist.append([np.array(temp1)])
    cls_dist = np.concatenate(cls_dist)
    cls_dist_recls = cls_dist.copy()
    mapping_recls = mapping.copy()
    tree_relationship = {}
    pending = set(pending)

    while(len(pending) > 1):
        (child_a, child_b) = divmod(np.argmax(cls_dist), cls_dist.shape[1])
        temp1 = [np.concatenate([[cls_dist[child_a]], [cls_dist[child_b]]]).max(axis=0)]
        cls_dist = np.concatenate([cls_dist, temp1], axis=0)
        temp1 = np.append(temp1, -1)
        temp1 = np.vstack(temp1)
        cls_dist = np.concatenate([cls_dist, temp1], axis=1)
        cls_dist = np.delete(cls_dist, [child_a, child_b], axis = 0)
        cls_dist = np.delete(cls_dist, [child_a, child_b], axis = 1)
        # change mapping
        cluster_a = mapping.inv[child_a]
        cluster_b = mapping.inv[child_b]  # cluster id
        tree_relationship[node_id] = (cluster_a, cluster_b)
        del mapping[cluster_a], mapping[cluster_b]
        pending.remove(cluster_a)
        pending.remove(cluster_b)
        for i in pending:
            if(mapping[i]>min([child_a, child_b]) and mapping[i]<max([child_a, child_b])):
                mapping[i] -= 1
            elif(mapping[i]>max([child_a, child_b])):
                mapping[i] -= 2
        mapping[node_id] = len(cls_dist)-1
        pending.add(node_id)
        node_id += 1

    # build tree structure
    pending = list(pending)
    tree = Tree()
    T0 = Node(identifier = pending[0])
    tree.add_node(T0)
    while(len(pending)>0):
        parent = pending[0]
        for i in tree_relationship[parent]:
            tree.add_node(Node(identifier = i), parent=parent)
            if(i in tree_relationship):
                pending.append(i)
        pending.remove(parent)

    # load depth info
    depths = {}
    depths_mapping = defaultdict(set)
    leaves = set(tree.leaves())
    for i in tree.all_nodes():
        depths[i] = tree.depth(node=i)
        if(i in leaves):
            depths_mapping[depths[i]].add(i)

    return cls_dist_recls, mapping_recls, tree, depths, depths_mapping


def extract_kmers(fna_i, fna_path, ksize, kmer_index_dict, kmer_index, Lv, spec, tree_dir, alpha_ratio, identifier):
    kmer_sta = defaultdict(int)
    for j in fna_i:
        for seq_record in SeqIO.parse(fna_path[j], "fasta"):
            temp = str(seq_record.seq)
            for k in range(0, len(temp)-ksize):
                forward = temp[k:k+ksize]
                reverse = seqpy.revcomp(forward)
                for kmer in [forward, reverse]:
                    try:
                        kmer_sta[kmer_index_dict[kmer]] += 1
                    except KeyError:
                        kmer_index_dict[kmer] = kmer_index
                        kmer_sta[kmer_index] += 1
                        kmer_index += 1
    alpha = len(fna_i) * alpha_ratio
    for x in kmer_sta:
        if(kmer_sta[x] >= alpha):
            Lv[identifier].add(x)
        else:
            spec[identifier].add(x)
    print(identifier, len(Lv[identifier]), len(spec[identifier]))
    return kmer_index


def get_leaf_union(depth, higher_union, depths_mapping, Lv, spec, leaf):
    union = depths_mapping[depth]-set([leaf])
    if(len(higher_union) == 0):
        res = set([])
        x = max(list(depths_mapping.keys()))
        for i in depths_mapping[x]:
            higher_union[x] = higher_union[x] | Lv[i.identifier] | spec[i.identifier]
        return get_leaf_union(depth, higher_union, depths_mapping, Lv, spec, leaf)
    elif(depth == max(list(depths_mapping.keys()))):
        res = set([])
    elif(depth+1 in higher_union):
        res = higher_union[depth+1]
    else:
        index = list(higher_union.keys())[0]
        higher_union[depth+1] = higher_union[index]
        diff_nodes = set([])
        for i in range(depth+1, index):
            diff_nodes = diff_nodes | depths_mapping[i]
        del higher_union[index]
        for i in diff_nodes:
            higher_union[depth+1] = higher_union[depth+1] | Lv[i.identifier] | spec[i.identifier]
        res = higher_union[depth+1]
    return res, union

def get_intersect(intersection, descendant_leaves, Lv, del_label, label): # (leaves, intersection)
    waitlist = descendant_leaves.copy()
    delete = []
    kmerset = []
    for i in intersection:
        if(intersection[i][0].issubset(descendant_leaves)):
            delete.append(i)
            waitlist = waitlist - intersection[i][0]
            kmerset.append(intersection[i][1])
    for i in delete:
        del intersection[i]
    if(len(kmerset) == 0):
        temp = list(descendant_leaves)
        kmer_t = Lv[temp[0]]
        for i in temp[1:]:
            kmer_t = kmer_t & Lv[i]
    elif(len(kmerset) == 1):
        temp = list(waitlist)
        kmer_t = kmerset[0]
        for i in temp:
            kmer_t = kmer_t & Lv[i]
    else:
        temp = list(waitlist)
        kmer_t = kmerset[0]
        for i in kmerset[1:]:
            kmer_t = kmer_t & i
        for i in temp:
            kmer_t = kmer_t & Lv[i]
    for i in waitlist:
        del_label[i][0] = 1
    intersection[label] = (descendant_leaves, kmer_t)
    return kmer_t


def get_diff(higher_union, descendant_leaves, depths, all_nodes, node, Lv, spec, del_label):
    diff = []
    for i in all_nodes:
        if(depths[i]==depths[node] and i!=node):
            if(i.identifier not in higher_union):
                delete = set([])
                waitlist = descendant_leaves[i.identifier].copy()
                kmerset = []
                for j in higher_union:
                    if(descendant_leaves[j].issubset(descendant_leaves[i.identifier])):
                        delete.add(j)
                        waitlist = waitlist - descendant_leaves[j]
                        kmerset.append(higher_union[j])
                waitlist = list(waitlist)
                if(len(kmerset) == 0):
                    kmer_t = Lv[waitlist[0]] | spec[waitlist[0]]
                    for j in waitlist[1:]:
                        kmer_t = kmer_t | Lv[j] | spec[j]
                elif(len(kmerset) == 1):
                    kmer_t = kmerset[0]
                    for j in waitlist:
                        kmer_t = kmer_t | Lv[j] | spec[j]
                else:
                    kmer_t = kmerset[0]
                    for j in kmerset[1:]:
                        kmer_t = kmer_t | j
                    for j in waitlist:
                        kmer_t = kmer_t | Lv[j] | spec[j]
                higher_union[i.identifier] = kmer_t
                for j in waitlist:
                    del_label[j][1] = 1
                for j in delete:
                    del higher_union[j]
            diff.append(higher_union[i.identifier])
    return diff


def delete(Lv, spec, del_label):
    de = []
    for i in del_label:
        if(del_label[i] == [1, 1]):
            del Lv[i], spec[i]
            de.append(i)
    for i in de:
        del del_label[i]


# build tree-based indexing structure
def build_tree(arg):
    # read parameters
    start = time.time()
    dist_matrix_file = arg[0]
    cls_file = arg[1]
    tree_dir = arg[2]
    ksize = arg[3]
    params = arg[4]
    alpha_ratio = params[0]
    minsize = params[1]
    maxsize = params[2]
    max_cls_size = params[3]

    # save genomes info
    fna_seq = bidict.bidict()    # : 1
    fna_path = {}

    # read dist matrix (represented by similarity: 1-dist)
    # output: dist, fna_path, fna_seq
    f = open(dist_matrix_file, "r")
    lines = f.readlines()
    f.close()
    index = 0
    d = lines[0].rstrip().split("\t")[1:]
    bac_label = 0
    for i in lines[0].rstrip().split("\t")[1:]:
        temp = i[i.rfind('/')+1:].split(".")[0]
        fna_seq[temp] = index
        fna_path[index] = i
        index += 1
    dist = []
    for line in lines[1:]:
        dist.append([np.array(list(map(float, line.rstrip().split("\t")[1:])))])
    dist = np.concatenate(dist)

    # read initial clustering results. fna_mapping, from 1 for indexing
    f = open(cls_file, 'r')
    lines = f.readlines()
    f.close()
    fna_mapping = defaultdict(set)
    for line in lines:
        temp = line.rstrip().split("\t")
        for i in temp[2].split(","):
            fna_mapping[int(temp[0])].add(fna_seq[i])

    # initially build tree
    cls_dist, mapping, tree, depths, depths_mapping = hierarchy(fna_mapping, dist)

    # initially extract k-mers
    kmer_index_dict = bidict.bidict()
    kmer_index = 1
    Lv = defaultdict(set)
    spec = defaultdict(set)    # k-mers <= alpha
    leaves = tree.leaves()
    for i in leaves:
        kmer_index = extract_kmers(fna_mapping[i.identifier], fna_path, ksize, kmer_index_dict, kmer_index, Lv, spec, tree_dir, alpha_ratio, i.identifier)
    end = time.time()
    print('- The total running time of k-mer extraction is ',str(end-start),' s\n')
    start = time.time()

    # leaf nodes check
    recls_label = 0

    leaves_check = []
    check_waitlist = reversed(leaves)
    while(True):
        if(recls_label):
            cls_dist, mapping, tree, depths, depths_mapping = hierarchy(fna_mapping, dist)
            leaves = tree.leaves()
            temp = {}
            temp2 = []
            for i in check_waitlist:
                if(i in fna_mapping):
                    temp2.append(i)
            check_waitlist = temp2.copy()
            for i in check_waitlist:
                temp[tree.get_node(i)] = depths[tree.get_node(i)]
            check_waitlist = []
            a = sorted(temp.items(), key=lambda x: x[1], reverse=True)
            for i in a:
                check_waitlist.append(i[0])
            for i in fna_mapping:
                if(i not in Lv):
                    kmer_index = extract_kmers(fna_mapping[i], fna_path, ksize, kmer_index_dict, kmer_index, Lv, spec, tree_dir, alpha_ratio, i)
        higher_union = defaultdict(set)
        for i in check_waitlist:
            diff, diff_nodes = get_leaf_union(depths[i], higher_union, depths_mapping, Lv, spec, i)
            kmer_t = Lv[i.identifier] - diff
            for j in diff_nodes:
                kmer_t = kmer_t - Lv[j.identifier]
            for j in diff_nodes:
                kmer_t = kmer_t - spec[j.identifier]
            print(str(i.identifier) + " checking", end = "\t")
            print(len(kmer_t))
            if(len(kmer_t) < minsize):
                leaves_check.append(i)
        if(len(leaves_check)>0):
            recls_label = 1
        else:
            break
        # re-clustering
        check_waitlist = []
        while(recls_label == 1):
            cluster_id = max(list(fna_mapping.keys())) + 1
            check_waitlist.append(cluster_id)
            leaf_a = leaves_check[0].identifier
            row_index = mapping[leaf_a]
            column_index = cls_dist[row_index].argmax()
            leaf_b = mapping.inv[column_index]  # (leaf_a, leaf_b)
            temp2 = fna_mapping[leaf_a] | fna_mapping[leaf_b]
            print(cluster_id, leaf_a, leaf_b, temp2)
            del fna_mapping[leaf_a], fna_mapping[leaf_b]
            if(leaf_a in Lv):
                del Lv[leaf_a], spec[leaf_a]
            if(leaf_b in Lv):
                del Lv[leaf_b], spec[leaf_b]
            del leaves_check[0]
            if(tree.get_node(leaf_b) in leaves_check):
                leaves_check.remove(tree.get_node(leaf_b))
            temp1 = [np.concatenate([[cls_dist[row_index]], [cls_dist[column_index]]]).max(axis=0)]
            cls_dist = np.concatenate([cls_dist, temp1], axis=0)
            temp1 = np.append(temp1, -1)
            temp1 = np.vstack(temp1)
            cls_dist = np.concatenate([cls_dist, temp1], axis=1)
            cls_dist = np.delete(cls_dist, [row_index, column_index], axis = 0)
            cls_dist = np.delete(cls_dist, [row_index, column_index], axis = 1)
            # change mapping
            del mapping[leaf_a], mapping[leaf_b]
            pending = list(fna_mapping.keys())
            pending.sort()
            for i in pending:
                if(mapping[i]>min([row_index, column_index]) and mapping[i]<max([row_index, column_index])):
                    mapping[i] -= 1
                elif(mapping[i]>max([row_index, column_index])):
                    mapping[i] -= 2
            fna_mapping[cluster_id] = temp2
            mapping[cluster_id] = len(cls_dist)-1
            if(len(leaves_check) == 0):
                break
    del higher_union

    # rebuild identifiers
    all_nodes = tree.all_nodes()
    all_leaves_id = set([])
    leaves = set(tree.leaves())
    for i in leaves:
        all_leaves_id.add(i.identifier)
    id_mapping = bidict.bidict()
    index = 1
    index_internal = len(leaves)+1
    for i in all_nodes:
        if(recls_label == 0):
            id_mapping[i.identifier] = i.identifier
        elif(i in leaves):
            id_mapping[i.identifier] = index
            index += 1
        else:
            id_mapping[i.identifier] = index_internal
            index_internal += 1
    leaves_identifier = list(range(1, len(leaves)+1))
    all_identifier = list(id_mapping.values())
    all_identifier.sort()

    # save2file
    f = open(tree_dir+"/tree_structure.txt", "w")
    os.system("mkdir "+tree_dir+"/kmers")
    os.system("mkdir "+tree_dir+"/overlapping_info")
    for nn in all_identifier:
        i = id_mapping.inv[nn]
        f.write("%d\t"%id_mapping[i])
        if(i == all_nodes[0].identifier):
            f.write("N\t")
        else:
            f.write("%d\t"%id_mapping[tree.parent(i).identifier])
        if(nn in leaves_identifier):
            f.write("N\t")
        else:
            [child_a, child_b] = tree.children(i)
            f.write("%d %d\t"%(id_mapping[child_a.identifier], id_mapping[child_b.identifier]))
        if(len(fna_mapping[i]) == 1):
            temp = list(fna_mapping[i])[0]
            temp = fna_seq.inv[temp]
            f.write("%s"%temp)
        f.write("\n")
    f.close()
    f = open(tree_dir+"/hclsMap_95_recls.txt", "w")
    for nn in leaves_identifier:
        i = id_mapping.inv[nn]
        f.write("%d\t%d\t"%(nn, len(fna_mapping[i])))
        temp1 = list(fna_mapping[i])
        for j in temp1:
            temp = fna_seq.inv[j]
            if(j == temp1[-1]):
                f.write("%s\n"%temp)
            else:
                f.write("%s,"%temp)
    f.close()
    end = time.time()
    print('- The total running time of re-clustering is ',str(end-start),' s\n')
    start = time.time()

    # build indexing structure
    kmerlist = set([])  # all kmers used
    length = {}
    overload_label = 0
    if(len(tree.leaves())>max_cls_size):
        overload_label = 1
    # from bottom to top (unique k-mers)
    uniq_temp = defaultdict(set)
    rebuilt_nodes = []
    descendant = defaultdict(set)  # including itself
    ancestor = defaultdict(set)
    descendant_leaves = defaultdict(set)
    ancestor[all_nodes[0].identifier].add(all_nodes[0].identifier)
    for i in all_nodes[1:]:
        ancestor[i.identifier] = ancestor[tree.parent(i.identifier).identifier].copy()
        ancestor[i.identifier].add(i.identifier)
    for i in reversed(all_nodes):
        print(str(id_mapping[i.identifier]) + " k-mer removing...")
        if(i in leaves):
            uniq_temp[i.identifier] = Lv[i.identifier]
            descendant_leaves[i.identifier].add(i.identifier)
        else:
            (child_a, child_b) = tree.children(i.identifier)
            descendant[i.identifier] = descendant[child_a.identifier] | descendant[child_b.identifier]
            descendant_leaves[i.identifier] = descendant_leaves[child_a.identifier] | descendant_leaves[child_b.identifier]
            uniq_temp[i.identifier] = uniq_temp[child_a.identifier] & uniq_temp[child_b.identifier]
            uniq_temp[child_a.identifier] = uniq_temp[child_a.identifier] - uniq_temp[i.identifier]
            uniq_temp[child_b.identifier] = uniq_temp[child_b.identifier] - uniq_temp[i.identifier]
        descendant[i.identifier].add(i.identifier)
    all_nodes_id = set(id_mapping.keys())
    # remove overlapping
    for i in reversed(all_nodes):
        print(str(id_mapping[i.identifier]) + " k-mer set building...")
        # no difference with sibling, subtree and ancestors
        if(i == all_nodes[0]):
            kmer_t = uniq_temp[i.identifier]
        else:
            diff = {}
            temp = all_nodes_id - descendant[i.identifier] - set([tree.siblings(i.identifier)[0].identifier]) - ancestor[i.identifier]
            for j in temp:
                diff[j] = len(uniq_temp[j])
            a = sorted(diff.items(), key=lambda x: x[1], reverse=True)
            kmer_t = uniq_temp[i.identifier]
            for j in a:
                k = j[0]
                kmer_t = kmer_t - uniq_temp[k]
            # remove special k-mers
            temp = all_leaves_id - descendant_leaves[i.identifier]
            diff = {}
            for j in temp:
                diff[j] = len(spec[j])
            a = sorted(diff.items(), key=lambda x: x[1], reverse=True)
            for j in a:
                k = j[0]
                kmer_t = kmer_t - spec[k]
        if(len(kmer_t) < minsize and overload_label==0):
            rebuilt_nodes.append(i)
            print("%d waiting for reconstruction..." % id_mapping[i.identifier])
        else:
            if(len(kmer_t) > maxsize):
                kmer_t = set(random.sample(kmer_t, maxsize))
            f = open(tree_dir+"/kmers/"+str(id_mapping[i.identifier]), "w")
            for j in kmer_t:
                f.write("%d "%j)
            f.close()
            length[i] = len(kmer_t)
            kmerlist = kmerlist | kmer_t
    del uniq_temp

    # rebuild nodes
    overlapping = defaultdict(dict)
    intersection = defaultdict(set)
    higher_union = defaultdict(set)
    del_label = {}
    for i in leaves:
        del_label[i.identifier] = [0, 0]
    for i in rebuilt_nodes:
        print(str(id_mapping[i.identifier]) + " k-mer set rebuilding...")
        kmer_t = get_intersect(intersection, descendant_leaves[i.identifier], Lv, del_label, i.identifier)
        diff = get_diff(higher_union, descendant_leaves, depths, all_nodes, i, Lv, spec, del_label)
        for j in diff:
            kmer_t = kmer_t - j
        lower_leaves = set([])
        for j in leaves:
            if(depths[j] < depths[i]):
                lower_leaves.add(j)
        if(len(kmer_t) > maxsize):
            kmer_overlapping_sta = defaultdict(int)
            for j in lower_leaves:
                kmer_o = Lv[j.identifier] & kmer_t
                for k in kmer_o:
                    kmer_overlapping_sta[k] += 1
            temp = sorted(kmer_overlapping_sta.items(), key=lambda kv:(kv[1], kv[0]))
            kmer_t = set([])
            for j in range(0, maxsize):
                kmer_t.add(temp[j][0])
        nkmer = {}
        f = open(tree_dir+"/kmers/"+str(id_mapping[i.identifier]), "w")
        index = 0
        for j in kmer_t:
            f.write("%d "%j)
            nkmer[j] = index
            index += 1
        length[i] = len(kmer_t)
        kmerlist = kmerlist | kmer_t
        # save overlapping info
        for j in lower_leaves:
            temp = Lv[j.identifier] & kmer_t
            if(len(temp)>0):
                ii = id_mapping[i.identifier]
                jj = id_mapping[j.identifier]
                overlapping[jj][ii] = set([])
                for k in temp:
                    overlapping[jj][ii].add(nkmer[k])
        delete(Lv, spec, del_label)

    for i in overlapping:
        f = open(tree_dir+"/overlapping_info/"+str(i), "w")
        f1 = open(tree_dir+"/overlapping_info/"+str(i)+"_supple", "w")
        count = -1
        for j in overlapping[i]:
            if(len(overlapping[i])!=0):
                f.write("%d\n"%j)
                for k in overlapping[i][j]:
                    f.write("%d "%k)
                f.write("\n")
                count += 2
                f1.write("%d %d\n"%(j, count))
        f.close()
        f1.close()

    # final saving
    f = open(tree_dir+"/reconstructed_nodes.txt", "w")
    for i in rebuilt_nodes:
        f.write("%d\n"%id_mapping[i.identifier])
    f.close()

    f = open(tree_dir+"/node_length.txt", "w")
    for nn in all_identifier:
        i = id_mapping.inv[nn]
        f.write("%d\t%d\n"%(nn, length[tree[i]]))
    f.close()

    kmer_mapping = {}
    index = 0
    f = open(tree_dir+"/kmer.fa", "w")
    for i in kmerlist:
        f.write(">1\n")
        f.write(kmer_index_dict.inv[i])
        kmer_mapping[i] = index
        index += 1
        f.write("\n")
    f.close()

    # change index
    files = os.listdir(tree_dir+"/kmers")
    for i in files:
        f = open(tree_dir+"/kmers/"+i, "r")
        lines = f.readlines()
        if(len(lines) == 0):
            continue
        d = lines[0].rstrip().split(" ")
        d = map(int, d)
        f = open(tree_dir+"/kmers/"+i, "w")
        for j in d:
            f.write("%d "%kmer_mapping[j])
        f.close()

    end = time.time()
    print('- The total running time of tree-based indexing struture building is ',str(end-start),' s\n')


def cal_cls_dist(dist, fna_i, fna_j):
    if(fna_i==fna_j):
        return -1
    temp = set([])
    for i in fna_i:
        for j in fna_j:
            temp.add(dist[i][j])
    return max(temp)


def get_difference(node, depths, all_nodes, descendant_leaves):
    diff = []
    depth_cutoff = depths[node]
    for i in all_nodes:
        if(depths[i]==depth_cutoff and i!=node):
            diff += descendant_leaves[i]
    return diff



import re
import os
from collections import defaultdict
from Bio import SeqIO
import seqpy
import scipy.sparse as sp
import numpy as np

def load_cls(cls95):
	d2=defaultdict(lambda:{}) # id -> {strains}
	d3={} # strain -> id
	f2=open(cls95,'r')
	#non_rep={}
	while True:
		line=f2.readline().strip()
		if not line:break
		ele=line.split('\t')
		strains=re.split(',',ele[-1])
		for s in strains:
			d2[int(ele[0])][s]=''
			d3[s]=int(ele[0])
	return d2,d3
	

def build_kmer_dict(ddir,k,dsi,cls_num):
	match_1=[]
	match_2=[]
	for i in range(cls_num):
		match_1.append(i+1)
		match_2.append('0')
	dlabel=defaultdict(lambda:{})
	#dseq=defaultdict(lambda:[]) # id -> [seq]
	#print(dsi)
	#exit()
	for pre in ddir:
		g=ddir[pre]
		seq_dict = {rec.id : rec.seq for rec in SeqIO.parse(g, "fasta")}
		for cl in seq_dict:
			seq=str(seq_dict[cl])
			#dseq[pre].append(seq)
			for i in range(len(seq)-k+1):
				kmer=seq[i:i+k]
				rev_kmer=seqpy.revcomp(seq[i:i+k])
				dlabel[kmer][dsi[pre]-1]=1
				dlabel[rev_kmer][dsi[pre]-1]=1
	return dlabel




def load_strains(infa_strains):
	ddir={} # pre -> strain fasta dir
	for filename in os.listdir(infa_strains):
		pre=re.split('\.',filename)[0]
		g=infa_strains+'/'+filename
		ddir[pre]=g
	return ddir

def build_cls_kmer(infa_strains,cls95):
	k=31
	did2s,ds2id=load_cls(cls95) # ID -> Strains {s1:'',s2:''}/ strain -> ID
	cls_num=len(did2s)
	ddir=load_strains(infa_strains)
	#dlabel=build_kmer_dict(ddir,k,ds2id,cls_num) # kmer -> {'c1':'','c2':'',...}
	return did2s,cls_num

def build_omatrix(infa_strains,cls95,cls_dir):
	import pickle
	did2s,cls_num=build_cls_kmer(infa_strains,cls95)
	for filename in os.listdir(cls_dir):
		cid=re.sub('C','',filename)
		cid=int(cid)
		if len(did2s[cid])>1:
			if not os.path.exists(cls_dir+'/'+filename+'/overlap_matrix.npz'):
				print(filename)
			continue
			kid=pickle.load(open(cls_dir+'/'+filename+'/all_kid.pkl','rb'))
			row=len(kid)
			column=len(range(cls_num))
			'''
			o=open(cls_dir+'/'+filename+'/overlap_matrix.csv','w+')
			head=[]
			for i in range(cls_num):
				head.append(str(i+1))
			o.write(','.join(head)+'\n')
			'''
			for k in kid:
				kid[k]=int(kid[k])
			mat=sp.dok_matrix((row,column), dtype=np.int8)
			#res=sorted(kid.items(),key=lambda d:d[1])
	
			for k in kid:
				mat[kid[k]-1,list(dlabel[k].keys())]=1
				'''
				r=r[0]
				outa=[dlabel[r][key] for key in sorted(dlabel[r].keys())]
				o.write(','.join(outa)+'\n')
				'''
			mat=mat.tocsr()
			sp.save_npz(cls_dir+'/'+filename+'/overlap_matrix.npz',mat)			
		print('C',cid,'\tfinished.')

	
build_omatrix('/home/yongxinji2/tem/Build_SDB/ref_Ecoil','../DB_Ecoil/Cluster_Result/hclsMap_95.txt','../DB_Ecoil/Kmer_Sets_L2/Kmer_Sets')
#build_omatrix('../Cutibacterium_acnes_small_pure','../Csmall_50/Cluster_Result/hclsMap_95.txt','../Csmall_50/Kmer_Sets_L2/Kmer_Sets')
#build_omatrix('../../Ref_Genome','../DB_Cae_20/Cluster_Result/hclsMap_95.txt','../DB_Cae_20/Kmer_Sets_L2/Kmer_Sets')

	



import re
import os
import Unique_kmer_detect_direct
import seqpy
from collections import defaultdict
from Bio import SeqIO
#import msa_polish_with_kalign
import math
import pickle
import Recls_withR_new
import scipy.sparse as sp
import numpy as np
import psutil
import gc

def build_dir(input_dir):
	if not os.path.exists(input_dir):
		os.makedirs(input_dir)

def unique_kmer_out_inside_cls(d,k,dlabel,out_dir):
	print('::Scan unique kmer inside cluster and output')
	count=1
	knum=1
	kid_match={} # Kmer -> ID
	sid_match={} # Strain -> ID
	ids_match={} # ID -> Strain
	head=[]
	#o=open(out_dir+'/unique_kmer_all.fasta','w+')
	pre_sim_d={}
	match_1=[]
	match_2=[]
	#all_uk={} # {uk1:'',uk2:'',....}
	for s in d:
		match_1.append(s)
		match_2.append('0')
		#pre_sim_d[s]='0' # strain id to 0 
	duniq_num=defaultdict(lambda:0)	
	kmatrix=defaultdict(lambda:{}) # {1:{1:0, 2:1, ....}} # Kmer id -> Strain id-> 0 or 1
	for s in d: # For each strain in the cluster -> 's' here refers to the id of strain
		head.append(s)
		count+=1
		uk_count=0
		resd={}
		for s2 in d[s]: # Only one time
			pre=Unique_kmer_detect_direct.get_pre(s2)
			sid_match[pre]=s # Name -> Strain id
			ids_match[s]=pre # ID -> Strain Name
			#o=open(out_dir+'/'+pre+'.fasta','w+')
			seq_dict = {rec.id : rec.seq for rec in SeqIO.parse(s2, "fasta")}
			for cl in seq_dict:
				seq=str(seq_dict[cl])
				#rev_seq=seqpy.revcomp(seq)
				for i in range(len(seq)-k+1):
					kmer=seq[i:i+k]
					if len(dlabel[kmer])==1:
						#duniq_num[s]+=1
						resd[kmer]=''
						resd[seqpy.revcomp(kmer)]=''
						

		kcount=0
		for kmr in resd:
			#uk_count+=1
			kcount+=1
			if kcount<10001:
				kid_match[kmr]=knum
				#head.append(str(knum))
				#kmatrix[knum]=dict(zip(match_1,match_2))
				kmatrix[knum][s-1]=1
				knum+=1
	
	tem=[]
	head=sorted(head)
	for h in head:
		tem.append(str(h))
	#o2.write(','.join(tem)+'\n')
	head_out=','.join(tem)+'\n'
	'''
	for kid in sorted(kmatrix.keys()):
		outa=[kmatrix[kid][key] for key in sorted(kmatrix[kid].keys())]
		#o2.write(','.join(outa)+'\n')
	#o2.close()

	with open(out_dir+'/uk_kid.pkl','wb') as o3:
		pickle.dump(kid_match, o3, pickle.HIGHEST_PROTOCOL)
	'''
	tem=[]
	for i in sorted(ids_match.keys()):
		tem.append(ids_match[i])
	with open(out_dir+'/id2strain.pkl','wb') as o4: # The list of strain name.
		pickle.dump(tem, o4, pickle.HIGHEST_PROTOCOL)
	#print('Unique part -> kid_match:',len(kid_match),', kmatrix:',len(kmatrix))
	return match_1,match_2,kid_match,kmatrix, head_out,knum,sid_match



	#return kmatrix,kid_match



def find_unique_kmers_inside_cls(d,out_dir,ksize,dlabel):
	#uk_dir=out_dir+'/Unique_Kmer'
	#cb_dir=out_dir+'/Colinear_Block'
	#ks_dir=out_dir+'/Kmer_Sets'
	#build_dir(uk_dir)
	#build_dir(cb_dir)
	#build_dir(ks_dir)
	'''
	if not os.path.exists(out_dir):
		os.makedirs(out_dir)
	'''
	temc=1
	temd=defaultdict(lambda:{})
	for e in d: # For each strain
		#if len(d[e])==1:continue
		# Find unique kmer for each strain in the clusters
		#temd=defaultdict(lambda:{})
		#temc=1
		temd[temc][e]=''
		'''
		for l in d[e]:
			temd[temc][l]=''
		'''
		temc+=1
		#dlabel=Unique_kmer_detect_direct.build_kmer_dict(temd,ksize)
		#uk_out=uk_dir+'/C'+str(e)
		#build_dir(uk_out)
	match_1,match_2,kid_match,kmatrix,head_out,knum,sid_match=unique_kmer_out_inside_cls(temd,ksize,dlabel,out_dir)
	return match_1,match_2,kid_match,kmatrix,head_out,knum,sid_match
	#return kmatrix,kid_match
	



def connect_genome(input_genome):
	seq_dict = {rec.id : rec.seq for rec in SeqIO.parse(input_genome, "fasta")}
	contig_num=len(seq_dict)
	all_seq=[]
	for s in seq_dict:
		seq=str(seq_dict[s])
		all_seq.append(seq)
	connect='N'*100
	out_seq=connect.join(all_seq)
	return out_seq,contig_num

def out_block(oi,dblock,block,strain_cls,mcgr):
	oi.write('a\n')
	block_seq=[]
	for s in dblock[block]:
		fs=open(strain_cls[s],'r')
		head=fs.readline()
		seq_all=fs.readline().strip()
		tem=[]
		tem.append('s')
		tem.append(s)
		tem.append(str(dblock[block][s][1]))
		seq=seq_all[dblock[block][s][1]:dblock[block][s][2]]
		if dblock[block][s][0]=='-':
			seq=seqpy.revcomp(seq)
		tem.append(str(len(seq)))
		tem.append(dblock[block][s][0])
		tem.append(str(len(seq_all)))
		tem.append(seq)
		tem_s=' '.join(tem)
		block_seq.append(tem_s)
		'''
		if s in mcgr:
			ref_match[dblock[block][s][1]]=[len(seq),dblock[block][s][0]]
			ref_match[-1]=len(seq_all)
		'''
	block_seq_out='\n'.join(block_seq)
	oi.write(block_seq_out+'\n\n')
	

def extract_unique_block_from_coords(block_coords,strain_cls,block_dir,mcgr):
	# With block coords from sibeliaz, we need to align these block using Kalign, so we extract them firstly
	f=open(block_coords,'r')
	dblock=defaultdict(lambda:{}) # bid -> 'strain_pre':[+,1811,2534] ->(strand, start, end)
	dblock_rep_check={} # bid ->  0 (without repetitive block) or 1 (with repetitive block)
	while True:
		line=f.readline().strip()
		if not line:break
		if line[0]=='#':continue
		ele=line.split()
		bid=re.sub('.*=','',ele[-1])
		if bid not in dblock_rep_check:
			dblock_rep_check[bid]=0
		if bid in dblock:
			if ele[0] in dblock[bid]:
				dblock_rep_check[bid]=1
		dblock[bid][ele[0]]=[ele[-3],int(ele[-6])-1,int(ele[-5])]
	o=open(block_dir+'/alignment_unique.maf','w+')
	o2=open(block_dir+'/alignment_global.maf','w+')
	#ref_match_p={} # One Ref genome -> Block Site and their length (-1 -> the total length of this strain)
	#ref_match_g={}
	for block in dblock:
		#if len(dblock[block])<int(len(strain_cls)*0.5) and dblock_rep_check[block]==0:
		#if len(dblock[block])<=int(len(strain_cls)*0.5):
		if not len(dblock[block])==len(strain_cls): # Use all
			# Partial colinear block
			out_block(o,dblock,block,strain_cls,mcgr)
		else:
			# Global colinear block
			out_block(o2,dblock,block,strain_cls,mcgr)

	

def split_arr(arr,m):
	'''
	for i in range(0, len(listTemp), n):
		yield listTemp[i:i + n]
	'''
	n = int(math.ceil(len(arr) / float(m)))
	return [arr[i:i + n] for i in range(0, len(arr), n)]

def load_split_block(input_block,split_num,out_dir):
	tem_dir=out_dir+'/Temd'
	build_dir(tem_dir)
	f=open(input_block,'r')
	lines=f.read().split('\n')
	blocks=[]
	c=0
	#block_match={}
	for l in lines:
		if not l:continue
		if l[0]=='a':
			if not c==0:
				tems='\n'.join(tema)
				tems='a\n'+tems
				tems=tems+'\n\n'
				blocks.append(tems)
				#block_match[c]=tems
			tema=[]
			c+=1
		else:
			tema.append(l)
	#block_match[c]=tems 
	blocks.append(tems)
	#print(len(blocks))
	#exit()
	#print(len(lines),lines[:2])
	sub_block=split_arr(blocks,split_num)
	#print(len(sub_block))
	#print(sub_block)
	#exit()
	back_arr=[] # [[block_dir,out_block,size],[...],...]
	c=1
	for s in sub_block:
		#s=s.strip()
		'''
		outa=[]
		for e in sub_block[s]:
			outa.append(e)
		'''
		outs=''.join(s)
		#outs=outs.strip()+'\n'
		sub_dir=tem_dir+'/B'+str(c)+'.maf'
		rebs_dir=tem_dir+'/B'+str(c)+'_rebuild.maf'
		o=open(sub_dir,'w+')
		o.write(outs)
		back_arr.append([sub_dir,rebs_dir,len(s)])
		c+=1
	return back_arr,tem_dir

def count_dbs(lines,ksize,alls):
	#d=defaultdict(lambda:{})
	d = defaultdict(lambda :defaultdict(defaultdict))
	d2={}
	c=1
	for line in lines:
		if not line:continue
		if line[0]=='a':
				blockid=c
				d2[blockid]=0
				c+=1
		if line[0]=='s':
			#ele=line.split()
			d2[blockid]+=1
	c=1
	for line in lines:
		if not line:continue
		if line[0]=='a':
			blockid=c
			c+=1
		if line[0]=='s':
			#print(d2[blockid],alls)
			if d2[blockid]>alls:continue
			ele=line.split()
			for i in range(len(ele[-1])-ksize+1):
				kmer=ele[-1][i:i+ksize]
				if not len(kmer)==ksize:continue
				if re.search('N',kmer):continue
				rev_kmer=seqpy.revcomp(kmer)
				d[blockid][kmer][ele[1]]=''
				d[blockid][rev_kmer][ele[1]]=''
	return d

def generate_kmer_match_from_global(input_gb,ksize,out_dir,dlabel,match_1,match_2,head_out,sid_match,label_match,knum,kid_match,kmatrix):
	f=open(input_gb,'r')
	lines=f.read().split('\n')
	o=open(out_dir+'/all_kmer.fasta','w+')
	#kmatrix=defaultdict(lambda:{}) # K-mer id -> Strain id -> '0' or '1'
	#kid_match={}
	#knum=1
	c=1
	#dk_match=defaultdict(lambda:{}) # Kmer ->  {strain1:'',strain2:''}
	#dbs_count=count_dbs(lines,ksize)
	dtotal_kmer={}
	for line in lines:
		if not line:continue
		if line[0]=='a':
			if c==1:
				#dstrain={}
				dtotal_kmer={}
				#blockid=c
				c+=1
			else:
				if len(dtotal_kmer)>0:
					for k in dtotal_kmer:
						if k in kid_match:continue
						if len(dlabel[k])==1:continue
						kid_match[k]=knum
						#kmatrix[knum]=dict(zip(match_1,match_2))
						for e in dlabel[k]:
							kmatrix[knum][sid_match[label_match[e]]-1]=1
						knum+=1
				#dstrain={}
				dtotal_kmer={}
				blockid=c
				c+=1
				
		if line[0]=='s':
			ele=line.split()
			#dstrain[ele[1]]=''
			for i in range(len(ele[-1])-ksize+1):
				kmer=ele[-1][i:i+ksize]
				if not len(kmer)==ksize:continue
				if re.search('N',kmer):continue
				rev_kmer=seqpy.revcomp(kmer)
				if not len(dlabel[kmer])==len(sid_match): 
					dtotal_kmer[kmer]=''
				if not len(dlabel[rev_kmer])==len(sid_match):
					dtotal_kmer[rev_kmer]=''
	if len(dtotal_kmer)>0:
		for k in dtotal_kmer:
			if k in kid_match:continue
			if len(dlabel[k])==1:continue
			kid_match[k]=knum
			#kmatrix[knum]=dict(zip(match_1,match_2))
			for e in dlabel[k]:
				kmatrix[knum][sid_match[label_match[e]]-1]=1
			knum+=1
	kc=1
	dlabel={}
	gc.collect()
	row=len(kmatrix)
	column=len(match_1)
	for nk in kid_match:
		o.write('>'+str(kc)+'\n'+nk+'\n')
		kc+=1
	print(u'Memory usage：%.4f GB' % (psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024 / 1024) )
	#o1.write(head_out)
	print('Fill the Sparse matrix: Row: ',row,' Column: ',column)
	mat=sp.dok_matrix((row,column),dtype=np.int8)
	print(u'Memory usage：%.4f GB' % (psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024 / 1024) )
	for kmr in kmatrix:
		mat[kmr-1,list(kmatrix[kmr].keys())]=1
	mat=mat.tocsr()
	sp.save_npz(out_dir+'/all_strains.npz',mat)

	with open(out_dir+'/all_kid.pkl','wb') as o2:
		pickle.dump(kid_match,o2,pickle.HIGHEST_PROTOCOL)
	# Now all sets are generated, we will re-cluster these strains to remove those 1% similar case
	Recls_withR_new.remove_1per(out_dir+'/all_strains.npz',out_dir+'/id2strain.pkl',out_dir)



def generate_kmer_match_from_uk(input_uk,ksize,out_dir,dlabel,match_1,match_2,head_out,knum,kid_match,kmatrix,sid_match):
	#import pickle
	f=open(input_uk,'r')
	lines=f.read().split('\n')
	#o=open(out_dir+'/partial_kmer_addUk2.fasta','w+')
	#kmatrix=defaultdict(lambda:{}) # K-mer id -> Strain id -> '0' or '1'
	#kid_match={}
	#knum=1
	#o2=open(out_dir+'/kmer_match.txt','w+')
	c=1
	#dk_match=defaultdict(lambda:{}) # Kmer ->  {strain1:'',strain2:''}
	alls=0.5*len(match_1)
	dbs_count=count_dbs(lines,ksize,alls)	# Block_ID -> {s1:'',s2:'',....}
	dtotal_kmer={}
	for line in lines:
		if not line:continue
		if line[0]=='a':
			if c==1:
				dstrain={}
				dtotal_kmer={}
				blockid=c
				c+=1
			else:
				#if len(dtotal_kmer)>(80-ksize+1)*2: # Minimum kmer cutoff -> Length: 100bp
				if len(dtotal_kmer)>0:
					if blockid in dbs_count:
						for k in dtotal_kmer:
							if len(dlabel[k])==1:continue
							if k in kid_match:continue
							kid_match[k]=knum
							for e in dict(dbs_count[blockid][k]):
								kmatrix[knum][sid_match[e]-1]=1
							knum+=1
					
							
				dstrain={}
				dtotal_kmer={}
				blockid=c
				c+=1

		if line[0]=='s':
			if blockid not in dbs_count:continue
			ele=line.split()
			dstrain[ele[1]]=''
			for i in range(len(ele[-1])-ksize+1):
				kmer=ele[-1][i:i+ksize]
				if not len(kmer)==ksize:continue
				if re.search('N',kmer):continue
				rev_kmer=seqpy.revcomp(kmer)

				if len(dlabel[kmer])==len(dbs_count[blockid][kmer]):
					dtotal_kmer[kmer]=''
				if len(dlabel[rev_kmer])==len(dbs_count[blockid][rev_kmer]):
					dtotal_kmer[rev_kmer]=''
	#if len(dtotal_kmer)>(80-ksize+1)*2:
	if len(dtotal_kmer)>0:
		#if blockid not in dbs_count:continue
		for k in dtotal_kmer:
			if blockid not in dbs_count:continue
			#dk_match[k][blockid]=dstrain
			#if len(dict(dbs_count[blockid][k]))==1:continue
			#dk_match[k]=dict(dbs_count[blockid][k])
			if len(dlabel[k])==1:continue
			if k in kid_match:continue
			kid_match[k]=knum 
			#kmatrix[knum]=dict(zip(match_1,match_2))
			for e in dict(dbs_count[blockid][k]):
				kmatrix[knum][sid_match[e]-1]=1
			knum+=1
	for k in kid_match:
		if k not in dlabel:continue
		if len(dlabel[k])==1:continue
		del dlabel[k]
	gc.collect()
	return knum,kid_match,kmatrix
	'''
	kc=1
	for nk in kid_match:
		o.write('>'+str(kc)+'\n'+nk+'\n')
		kc+=1
	
	
	with open(out_dir+'/kmatch.pkl', 'wb') as o2:
		pickle.dump(dict(dk_match), o2, pickle.HIGHEST_PROTOCOL)
	
	
	o1=open(out_dir+'/pk_uk2_strain.csv','w+')
	o1.write(head_out)
	for kid in sorted(kmatrix.keys()):
		outa=[kmatrix[kid][key] for key in sorted(kmatrix[kid].keys())]
		o1.write(','.join(outa)+'\n')
	with open(out_dir+'/pk_uk2_kid.pkl', 'wb') as o2:
		pickle.dump(kid_match, o2, pickle.HIGHEST_PROTOCOL)
	'''



def build_kmer_dict(d,k):
	print('Load kmer to dict...')
	import time
	dlabel=defaultdict(lambda:{})
	c=1
	label_match={}
	for g in d:
		print('Process: ',c,'/',len(d))
		seq_dict = {rec.id : rec.seq for rec in SeqIO.parse(g, "fasta")}
		for cl in seq_dict:
			seq=str(seq_dict[cl])
			#rev_seq=seqpy.revcomp(seq)
			for i in range(len(seq)-k+1):
				kmer=seq[i:i+k]
				rev_kmer=seqpy.revcomp(seq[i:i+k])
				#rev_kmer=rev_seq[i:i+k]
				dlabel[kmer][c]=''
				dlabel[rev_kmer][c]=''
		pre=Unique_kmer_detect_direct.get_pre(g)
		label_match[c]=pre
		c+=1
	return dlabel,label_match


def build_kmer_sets(d,out_dir,ksize):
	print('Now we will extract kmers from unique region found by sibeliaz')
	#import multiprocessing
	ksize=int(ksize)
	cb_dir=out_dir+'/Colinear_Block'
	ks_dir=out_dir+'/Kmer_Sets'
	#uk_region_dir=out_dir+'/Unique_Region'
	build_dir(cb_dir)
	build_dir(ks_dir)
	#build_dir(uk_region_dir)
	for e in d: # Go into a cluster
		#if len(d[e])==1 or len(d[e])==2:continue
		if len(d[e])==1:
			continue
		if len(d[e])<1000:continue # Just for test, need to annotate later
		cb_out=cb_dir+'/C'+str(e)
		cg_dir=cb_out+'/Connect_Genomes'
		matrix_out=ks_dir+'/C'+str(e)
		build_dir(cg_dir)
		build_dir(matrix_out)
		strain_cls={} # Pre -> Connect genome dir
		strains=[]
		mcgr={} # most_complete_genome_random: genome -> contig number
		min_contig=-1
		# Build kmer -> Strain number dict
		dlabel,label_match=build_kmer_dict(d[e],int(ksize))
		#dlabel=klabel[int(e)]
		#label_match=mapping
		#print(dlabel)
		for s in d[e]: # Each strain inside the cluster
			pre=Unique_kmer_detect_direct.get_pre(s)
			cg_name=cg_dir+'/'+pre+'.fasta'
			strain_cls[pre]=cg_name
			strains.append(cg_name)
			o=open(cg_name,'w+')
			connect_seq,contig_num=connect_genome(s)
			o.write('>'+pre+'\n'+connect_seq+'\n')
			# Take the genome with minimum contig as reference
			if len(mcgr)==0:
				mcgr[pre]=contig_num
				min_contig=contig_num
			if contig_num<min_contig:
				mcgr={}
				mcgr[pre]=contig_num 
				min_contig=contig_num
		# Use sibeliaz to obtain colinear block
		block_dir=cb_out+'/Blocks'
		build_dir(block_dir)
		# Run sibeliaz to find local colinear blocks
		all_s=' '.join(strains)

		#os.system('sibeliaz -n -m 100 -k 15 -a '+str(150*len(strain_cls))+' -o '+block_dir+' '+all_s)

		# Extract Partial colinear blocks from these blocks 
		#extract_unique_block_from_coords(block_dir+'/blocks_coords.gff',strain_cls,block_dir,mcgr)
		
		#exit()
		#uk_cls=uk_region_dir+'/C'+str(e)
		#build_dir(uk_cls)

		# K-mer matrix from unique k-mers
		match_1,match_2,kid_match,kmatrix,head_out,knum,sid_match=find_unique_kmers_inside_cls(d[e],matrix_out,ksize,dlabel)
		print('Unique part -> kid_match:',len(kid_match),', kmatrix:',len(kmatrix))
		# K-mer matrix from partial colinear block
		knum,kid_match,kmatrix=generate_kmer_match_from_uk(block_dir+'/alignment_unique.maf',int(ksize),matrix_out,dlabel,match_1,match_2,head_out,knum,kid_match,kmatrix,sid_match)
		print('Partial Unique part -> kid_match:',len(kid_match),', kmatrix:',len(kmatrix))
		# K-mer matrix from global colinear block - Only mode/ All mode
		generate_kmer_match_from_global(block_dir+'/alignment_global.maf',int(ksize),matrix_out,dlabel,match_1,match_2,head_out,sid_match,label_match,knum,kid_match,kmatrix)

		
		#find_unique_kmers_inside_cls(d[e],matrix_out,ksize,dlabel)
		#exit()
		# Finished 	
		#return block_dir,mcgr,ref_match # This line shows -> What we can obtain in this step
		# Continue: Align sequences with kalign
		# Without multiprocess to speed up
		
		#input_block=block_dir+'/alignment.maf'
		#out_block=block_dir+'/alignment_rebuild.maf'
		'''
		msa_polish_with_kalign.realign_with_kalign([input_block,out_block,len(ref_match)-1])
		'''
		# Add 2021-02-01 Use multiprocess to speed up
		# Basic logic: Split alignment file and align them sperately
		
		#back_arr,tem_dir=load_split_block(input_block,4,block_dir)
		#print('Spilt_finished....')
		#msa_polish_with_kalign.realign_with_kalign(back_arr[0])
		#exit()
		'''
		pool = multiprocessing.Pool(processes = 4)
		group=1
		for item in back_arr:
			print('::Go:','Group ',group)
			group+=1
			pool.apply_async(msa_polish_with_kalign.realign_with_kalign,(item,))
		pool.close()
		pool.join()
		'''
